"""
SimPype's simulation.

.. code-block :: python

	# Import SimPype module
	import simpype
	# Import python random module
	import random

	# [Mandatory] Create a SimPype simulation object
	sim = simpype.Simulation(id = 'simple')
	# [Optional] Fix the seed for the pseudo-random generator
	sim.seed = 42
	# [Optional] Configure the log directory. 
	# [Default] Log are stored by default in the 'current working directory/log'
	sim.log.dir = 'mylog'

	# [Mandatory] Add at least one generator to the simulation
	gen0 = sim.add_generator(id = 'gen0')
	# [Mandatory] Assign an arrival time
	# Generator.random is a custom dictionary accepting the following format as values:
	# generator.random[<some_id>] = {
	# 	<initial_time> : lambda: <value>/<random_function>
	#	...
	# }
	# Random values can be generated in the following way:
	# 	generator.random[<some_id>].value
	# The random value is:
	# 	<value>/<random_function> the simulation time is equal or 
	# greater than (>=) <initial_time>, 0 otherwise
	gen0.random['arrival'] = {
		# From t=0 to t=10, arrival is constant every 3s
		0	: lambda: 3.0,
		# From t=10 to t=20, arrival is uniform between 2.5 and 3.5
		10	: lambda: random.uniform(2.5, 3.5),
		# From t=20 to t=inf, arrival is expovariate with lambda 0.20
		20	: lambda: random.expovariate(0.20)
	}

	# [Mandatory] Add at least one resource to the simulation
	res0 = sim.add_resource(id = 'res0')
	# [Mandatory] Assign a service time
	# Resource.random is a dictionary accepting the same Generator.random format
	res0.random['service'] = {
		# From t=0 to t=10, service is constant at 1.5s
		0	: lambda: 1.5,
		# From t=10 to t=20, service is uniform between 1.5 and 2.5
		10	: lambda: random.uniform(1.5, 2.5),
		# From t=20 to t=inf, arrival is expovariate with lambda 2.0
		20	: lambda: random.expovariate(2.0)
	}

	# [Mandatory] Add a pipeline connecting the generator and the resource
	p0 = sim.add_pipeline(gen0, res0)

	# [Mandatory] Run the simulation e.g. until t=30
	#             sim.run calls Simpy's env.run
	#             Any arg passed to sim.run is then passed to env.run
	sim.run(until = 30)


The log directory structure is the following:

.. code-block :: none

	log.dir
	|-- <simulation #1>
	|   |-- <run #1>
	|   |   |-- sim.cfg
	|   |   `-- sim.log
	|   |-- <run #2>
	|   |   |-- sim.cfg
	|   |   `-- sim.log
	|-- <simulation #2>
	|   |-- <run #1>
	|   |   |-- sim.cfg
	|   |   `-- sim.log
	|   |-- <run #2>
	|   |   |-- sim.cfg
	|   |   `-- sim.log
	|   ...
	...


``sim.cfg`` contains information about the simulation environment and has the following format:

.. code-block :: none

	Simulation Seed: 42
	Simulation Time: 30.000000000
	Execution Time: 0.003298451

``sim.log`` contains the actual log of the simulation events and has the following format:

.. code-block :: none

	timestamp,message,seq_num,resource,event
	0.000000000,gen0,0,res0,pipe.in
	0.000000000,gen0,0,res0,pipe.out
	1.500000000,gen0,0,res0,resource.serve
	3.000000000,gen0,1,res0,pipe.in
	3.000000000,gen0,1,res0,pipe.out
	4.500000000,gen0,1,res0,resource.serve
	6.000000000,gen0,2,res0,pipe.in
	6.000000000,gen0,2,res0,pipe.out
	7.500000000,gen0,2,res0,resource.serve
	9.000000000,gen0,3,res0,pipe.in
	9.000000000,gen0,3,res0,pipe.out
	10.500000000,gen0,3,res0,resource.serve
	12.000000000,gen0,4,res0,pipe.in
	12.000000000,gen0,4,res0,pipe.out
	13.525010755,gen0,4,res0,resource.serve
	15.139426798,gen0,5,res0,pipe.in
	15.139426798,gen0,5,res0,pipe.out
	16.862637537,gen0,5,res0,resource.serve
	17.914456117,gen0,6,res0,pipe.in
	17.914456117,gen0,6,res0,pipe.out
	20.091155604,gen0,6,res0,resource.serve
	21.150927331,gen0,7,res0,pipe.in
	21.150927331,gen0,7,res0,pipe.out
	21.196403533,gen0,7,res0,resource.serve

"""
import copy
import datetime
import functools
import inspect
import os
import random
import shutil
import simpy
import time

import simpype
import simpype.build


class Model:
	""" Class storing the simulation parameters regarding the dynamic models.

	Args:
		sim (:class:`Simulation`):
			The SimPype simulation object.
	
	Attributes:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		env (simpy.Environment):
			The SimPy environment object.

	"""
	def __init__(self, sim):
		assert isinstance(sim, Simulation)
		self.sim = sim
		self.env = sim.env
		self.dir = os.getcwd()
	
	@property
	def dir(self):
		""" The folder containing the customs models to be loaded by the SimPype simulation environment. """
		return self._dir

	@dir.setter
	def dir(self, val):
		self._dir = os.path.abspath(val)


class Log:
	""" Class storing the simulation parameters regarding the dynamic models.

	Args:
		sim (:class:`Simulation`):
			The SimPype simulation object.
	
	Attributes:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		env (simpy.Environment):
			The SimPy environment object.
		date (datetime.now()):
			The real world creation time of the simulation environment.
		file (bool):
			Write the logs to a file if ``True``. Default value is ``True``.
		print(bool):
			Print the logs to the console if ``True``. Default value is ``False``.

	"""
	def __init__(self, sim):
		assert isinstance(sim, Simulation)
		self.sim = sim
		self.env = sim.env
		self.date = datetime.datetime.now()
		self.dir = os.path.join(os.getcwd(), 'log')
		self.file = True
		self.print = False
		self._h_fixed = ["timestamp", "message", "seq_num", "resource", "event"]
		self._h_property = []
		self._first = True

	@property
	def dir(self):
		""" The folder where the simulation logs are written. """
		return self._dir

	@dir.setter
	def dir(self, val):
		self._dir = functools.reduce(os.path.join,[val, self.sim.id, str(self.date)])

	def _write_log(self, timestamp):
		assert isinstance(timestamp, simpype.message.Timestamp)
		s = ""
		if self._first:
			s = s + ",".join(self._h_fixed + self._h_property) + "\n"
			self._first = False
		s = s + "%.9f" % timestamp.timestamp + "," + timestamp.message.id + "," + \
			str(timestamp.message.seq_num) + "," + timestamp.resource.id + "," + timestamp.description
		for h in self._h_property:
			if h not in timestamp.message.property:
				timestamp.message.property[h] = 'NA'
			p = timestamp.message.property[h]
			p = "%.9f" % p if isinstance(p, float) else str(p.value)
			s = s + "," + p
		if self.file:
			self._log.info(s)
		if self.print:
			print(s)
	
	def _write_cfg(self, entry):
		if self.file:
			self._cfg.info(str(entry))
		if self.print:
			print(str(entry))

	def init(self):
		""" Initialize the log folder and the simulation loggers. """
		if self.file:
			if not os.path.exists(self.dir):
				os.makedirs(self.dir)
			self._log = simpype.build.logger('log', os.path.join(self.dir, 'sim.log'))
			self._cfg = simpype.build.logger('cfg', os.path.join(self.dir, 'sim.cfg')) 

	def write(self, entry):
		""" Write a log entry.
			
		Args:
			entry (:class:`~simpype.message.Timestamp`)(str):
				The entry to be logged. If the entry is :class:`~simpype.message.Timestamp`

		"""
		if isinstance(entry, simpype.message.Timestamp):
			self._write_log(entry)
		else:
			self._write_cfg(entry)

	def property(self, property):
		""" Enable the logging of a message property.

		Args:
			property (str):
				The property to log.

		"""
		self._h_property.append(property)


class Simulation:
	""" Class implementing the SimPype's simulation environment.

	Args:
		id (str):
			The simulation environment id.
	
	Attributes:
		env (simpy.Environment):
			The SimPy environment object.
		id (str):
			The simulation environment id.
		resource (dict):
			The dictionary storing all the :class:`~simpype.resource.Resource` objects of the simulation.
		generator (dict):
			The dictionary storing all the :class:`~simpype.resource.Resource` objects implementing generator functionalities of the simulation.
		pipeline (dict):
			The dictionary storing all the :class:`~simpype.pipeline.Pipeline` objects of the simulation.
		log (:class:`Log`):
			The log class storing the logging parameters.
		model (:class:`Model`):
			The model class storing the custom model parameters.

	"""
	def __init__(self, id):
		self.env = simpy.Environment()
		self.id = id
		self.resource = {}
		self.generator = {}
		self.pipeline = {}
		self.seed = hash(random.random())
		self.log = Log(self)
		self.model = Model(self)

	@property
	def seed(self):
		""" The seed of the pseudo-random number generator used by this simulation. """
		return self._seed
	
	@seed.setter
	def seed(self, num):
		self._seed = num
		random.seed(num)

	def _update_message(self, pipeline):
		# Add the pipeline to the messages in the new pipeline
		generator = set(pipeline.resource.keys()) & set(self.generator.keys())
		for id in generator:
			self.generator[id].message.pipeline = pipeline

	def add_generator(self, id, model = None):
		""" Add a :class:`~simpype.resource.Resource` implementing generator funcionalities to the simulation environment.

		Args:
			id (str):
				The generator id
			model (str):
				The model of the generator. If model is ``None``, the default model is used.

		Returns:
			:class:`~simpype.resource.Resource`

		"""
		assert id not in self.generator
		assert id not in self.resource
		generator = simpype.build.generator(self, id, model)
		self.generator[generator.id] = generator
		self.resource[generator.id] = generator
		return generator

	def add_pipeline(self, *args):
		""" Chain multiples :class:`~simpype.resource.Resource` or :class:`~simpype.pipeline.Pipeline` objects into a pipeline..

		Args:
			*args (:class:`~simpype.resource.Resource`)(:class:`~simpype.pipeline.Pipeline`):
				Create a pipeline by chaining 2 or more :class:`~simpype.resource.Resource` or :class:`~simpype.pipeline.Pipeline` objects.
		
		Returns:
			:class:`~simpype.pipeline.Pipeline`

		"""
		assert len(args) > 1
		id = len(self.pipeline)
		pipeline = simpype.Pipeline(self, id)
		for i in range(0, len(args)-1):
			src, dst = args[i], args[i+1]
			pipeline.add_pipe(src, dst)
		self._update_message(pipeline)
		self.pipeline[pipeline.id] = pipeline
		return pipeline
	
	def add_resource(self, id, model = None, capacity = 1, pipe = None):
		""" Add a :class:`~simpype.resource.Resource` object to the simulation environment. 
		
		Args:
			id (str):
				The resource id
			model (str):
				The model of the resource. If model is ``None``, the default model is used.
			capacity (int):
				The capacity of the resource, that is the number of :class:`~simpype.message.Message` objects that the resource can simultaneously serve.
			pipe (str):
				The model of the pipe associated to the resource. If model is ``None``, the default model is used.

		Returns:
			:class:`~simpype.resource.Resource`

		"""
		assert id not in self.resource
		resource = simpype.build.resource(self, id, model, capacity, pipe)
		self.resource[resource.id] = resource
		return resource

	def merge_pipeline(self, *args):
		""" Merge varioues :class:`~simpype.pipeline.Pipeline` objects into a single pipeline.

		Args:
			*args (:class:`~simpype.pipeline.Pipeline`):
				Create a pipeline by merging 2 or more :class:`~simpype.pipeline.Pipeline` objects.
		
		Returns:
			:class:`~simpype.pipeline.Pipeline`

		"""
		assert len(args) > 1
		id = len(self.pipeline)
		pipeline = simpype.Pipeline(self, id)
		for i in range(0, len(args)):
			pipeline.merge_pipe(args[i])
		self._update_message(pipeline)
		self.pipeline[pipeline.id] = pipeline
		return pipeline

	def run(self, *args, **kwargs):
		""" Run the simulation environment using SimPy environment. """
		self.log.init()
		sptime = time.process_time()
		self.env.run(*args, **kwargs)
		eptime = time.process_time()
		# Save some simulation parameters
		self.log.write("Simulation Seed: "+ str(self.seed))
		self.log.write("Simulation Time: " + "%.9f" % self.env.now)
		self.log.write("Execution Time: " + "%.9f" % (eptime - sptime))



"""
SimPype's pipeline.

"""

import simpype.resource
import simpype.simulation


class Pipeline:
	""" The pipeline connecting the various :class:`~simpype.resource.Resource` instances.

	Args:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		id (str):
			The pipeline id.

	Attributes:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		env (simpy.Environment):
			The SimPy environment object.
		id (str):
			The pipeline id.
		resource (dict):
			The dictionary storing the connected resources in the form of adjency list.
		first (:class:`~simpype.resource.Resource`):
			The first resource of the pipeline.
		last (:class:`~simpype.resource.Resource`):
			The last resource of the pipeline.

	"""
	def __init__(self, sim, id):
		assert isinstance(sim, simpype.simulation.Simulation)
		self.sim = sim
		self.env = sim.env
		self.id = id
		self.resource = {}
		self.first = None
		self.last = None

	def add_pipe(self, src, dst):
		""" Add a pipe to the pipeline.
		
		Connect resource ``src`` to resource ``dst``.

		Args:
			src (:class:`~simpype.resource.Resource`):
				The source resource of the pipe.
			dst (:class:`~simpype.resource.Resource`):
				The destination resource of the pipe.

		"""
		assert isinstance(src, (simpype.resource.Resource, Pipeline))
		assert isinstance(dst, (simpype.resource.Resource, Pipeline))
		# Src check
		tsrc = src if isinstance(src, simpype.resource.Resource) else src.last
		# Dst check
		tdst = dst if isinstance(dst, simpype.resource.Resource) else dst.first
		# Create the list in the dictionary if does not exist
		if tsrc.id not in self.resource:
			self.resource[tsrc.id] = []
		# Add the element to the pipe dictionary
		self.resource[tsrc.id].append(tdst)
		# Merge the pipe if src is pipeline instance
		if isinstance(src, Pipeline):
			self.merge_pipe(src)
		# Merge the pipe if dst is pipeline instance
		if isinstance(dst, Pipeline):
			self.merge_pipe(dst)
		# Set the first element of the pipe
		if self.first is None:
			self.first = src if isinstance(src, simpype.resource.Resource) else src.first
		# Set the last element of the pipe
		self.last = dst if isinstance(dst, simpype.resource.Resource) else dst.last

	def merge_pipe(self, pipeline):
		""" Merge ``pipeline`` with this pipeline.

		Args:
			pipeline (:class:`Pipeline`):
				The pipeline to be merged with this pipeline.

		"""
		assert isinstance(pipeline, Pipeline)
		for key, value in pipeline.resource.items():
			if key not in self.resource:
				self.resource[key] = []
			self.resource[key] = self.resource[key] + list(set(value)-set(self.resource[key]))



"""
:class:`~simpype.random.Random` is a custom dictionary accepting the following format as value:

.. code-block :: python

	sim = simpype.Simulation(id = 'test')
	myrand = simpype.Random(sim, {
		initial_time : lambda_function
		...
	})

Where each dictionary element is so defined:

	* ``initial_time`` is the element key and must be of *int* or *float* type. It represents the initial simulation time at which the ``lambda_function`` is invoked;
	* ``lambda_function`` is the element value. It is mandatory that for the value to be a *lambda* function. Such function must return a value, usually a *int* or a *float*;

An example of random dictionary initialization is the following:

.. code-block :: python

	sim = simpype.Simulation(id = 'test')
	myrand = simpype.Random(sim, {
		# From t=0 to t=10, the random variable returns 
		# the constant value of 3.0
		0	: lambda: 3.0,
		# From t=10 to t=20, the random variable returns 
		# value uniformly distributed between 2.5 and 3.5
		10	: lambda: random.uniform(2.5, 3.5),
		# From t=20 to t=inf, the random variable returns 
		# a value exponentially distributed with lambda 0.20
		20	: lambda: random.expovariate(0.20)
	})

A second example of random dictionary with generation interrupts is the following:

.. code-block :: python

	sim = simpype.Simulation(id = 'test')
	myrand = simpype.Random(sim, {
		# From t=0 to t=10, the random variable returns 
		# the constant value of 3.0 after time 10.
		# From time 0 to 10, no random variable is generated
		10	: lambda: 3.0,
		# From t=10 to t=20, no random variable is generated
		10	: lambda: None,
		# From t=20 to t=30, the random variable returns 
		# a value exponentially distributed with lambda 0.20
		20	: lambda: random.expovariate(0.20)
		# From t=30 to t=inf, no random variable is generated
		30	: lambda: None
	})

Produce a random value:

.. code-block :: python

	# Simulation time = 5.0
	random_value = myrand.value    # random_value = 3.0
	...
	# Simulation time = 15.0
	random_value = myrand.value    # random_value = 3.2476115513945767
	...
	# Simulation time = 25.0
	random_value = myrand.value    # random_value = 7.374759019459148

"""
import random

import simpype


class Random:
	""" SimPype's random class that may return different values depending on the simulation time.

	Args:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		step_dict (dict):
			The dictionary storing the random steps.

	Attributes:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		env (simpy.Environment):
			The SimPy environment object.
		step_dict (dict):
			The dictionary storing the random steps.
		step_list (list):
			The list storing the sorted random steps.
	
	"""
	class Step:
		def __init__(self, tfrom, tto, process):
			self.tfrom = tfrom
			self.tto = tto
			self.process = process

	def __init__(self, sim, step_dict):
		assert isinstance(sim, simpype.Simulation)
		self.sim = sim
		self.env = sim.env
		self.step_dict = {}
		self.step_list = []
		# Init	
		for v in step_dict.values():
			assert callable(v)

		if 0 not in step_dict:
			step_dict[0] = lambda: None
		step_dict[float("inf")] = lambda: None
		time_step = sorted(step_dict)

		# Parse and build the steps
		i = 0
		while i < len(time_step)-1:
			tfrom = time_step[i]

			j = i+1
			while j < len(time_step):
				tto = time_step[j]
				a = step_dict[time_step[i]]
				b = step_dict[time_step[j]]

				create = True
				if a() is not None:
					process = a
					create = True
					# break
					i = i+1
					j = float("inf")
				elif a() is None and b() is not None:
					wait = (tto - tfrom) + b()
					process = lambda: wait
					create = True
					# break
					i = j
					j = float("inf")
				elif a() is None and b() is None:
					if j == len(time_step) - 1:
						process = b
						create = True
						# break
						i = float("inf")
						j = float("inf")
					else:
						j = j+1
					
				if create:
					s = self.Step(tfrom, tto, process)
					self.step_list.append(s)
					self.step_dict[s.tfrom] = s

	@property
	def value(self):
		""" Returns a random value given the current simulation time.

		Returns:
			Value as returned by the the ``lambda`` function.
			
		"""
		while self.step_list[0].tto <= self.env.now:
			self.step_list.pop(0)
		return self.step_list[0].process()


class RandomDict(dict):
	""" A custom dictionary storing :class:`Random` objects.

	Args:
		sim (:class:`Simulation`):
			The SimPype simulation object.

	Attributes:
		sim (:class:`Simulation`):
			The SimPype simulation object.
	
	"""
	def __init__(self, sim):
		assert isinstance(sim, simpype.Simulation)
		super().__init__()
		self.sim = sim
		self.env = sim.env
	
	def __setitem__(self, key, value):
		if isinstance(value, Random):
			super().__setitem__(key, value)
		else:
			super().__setitem__(key, Random(self.sim, value))



import simpype


class Pipe(simpype.Pipe):
	def __init__(self, sim, resource, id):
		super().__init__(sim, resource, id)
		self.add_queue(id = 'default')

	@simpype.pipe.dequeue
	def dequeue(self):
		return self.queue['default'].pop()
						        
	@simpype.pipe.enqueue
	def enqueue(self, message):
		return self.queue['default'].push(message)
															

# Do NOT remove
pipe = lambda *args: Pipe(*args)



import simpype


class Queue(simpype.Queue):
	def __init__(self, sim, pipe, id):
		super().__init__(sim, pipe, id)

	@simpype.queue.push
	def push(self, message):
		assert isinstance(message, simpype.Message)
		if len(self.buffer) < self.capacity:
			self.buffer.append(message)
			return message
		else:
			message.drop('full')
			return None

	@simpype.queue.pop
	def pop(self):
		if len(self.buffer) > 0:
			return self.buffer.pop(0)
		else:
			return None

# Do NOT remove
queue = lambda *args: Queue(*args)



"""
SimPype's queue.

"""
import types

import simpype


def __push(func, queue, message):
	assert isinstance(queue, simpype.Queue)
	assert isinstance(message, simpype.Message)
	message.location = queue
	message.timestamp('pipe.in')
	result = func(queue, message)
	if isinstance(result, simpype.Message) and queue.active.triggered:
		queue.pipe.full()
	return result

def __pop(func, queue):
	assert isinstance(queue, Queue)
	message = func(queue)
	if isinstance(message, simpype.Message):
		message.timestamp('pipe.out')
		message.queue = None
	return message

def push(arg):
	""" Decorator for overloading the default :class:`Queue` push behavior.

	Args:
		arg (:class:`Queue`)(``self``):
			The :class:`Queue` instance.
			
	If the overloading is done in scripts, the :class:`Queue` instance must be provided as decorator argument.
	
	.. code-block:: python

		myresource = sim.add_resource(id = 'myresource')
		myresource.add_queue(id = 'myqueue')

		@simpype.queue.push(myresource.pipe['myqueue'])
		def push(self, message):
			self.buffer.append(message)
			return message

	If the overloading is done inside a Queue subclass, the decorator must be called without any arguments. :class:`Queue` instance is automatically provided through ``self``.

	.. code-block:: python
	   
		class MyQueue(simpype.Queue):
			def __init__(self, sim, pipe, id):
				super().__init__(sim, pipe, id)

			@simpype.queue.push
			def push(self, message):
				self.buffer.append(message)
				return message
	
	"""
	if isinstance(arg, simpype.Queue):
		queue = arg
		def decorator(func):
			def wrapper(queue, message):
				return __push(func, queue, message)
			queue.push = types.MethodType(wrapper, queue)
			return wrapper
		return decorator
	else:
		func = arg
		def wrapper(queue, message):
			return __push(func, queue, message)
		return wrapper

def pop(arg):
	""" Decorator for overloading the default :class:`Queue` pop behavior.

	Args:
		arg (:class:`Queue`)(``self``):
			The :class:`Queue` instance.
			
	If the overloading is done in scripts, the :class:`Queue` instance must be provided as decorator argument.
	
	.. code-block:: python

		myresource = sim.add_resource(id = 'myresource')
		myresource.add_queue(id = 'myqueue')

		@simpype.queue.pop(myresource.pipe['myqueue'])
		def pop(self):
			return self.buffer.pop(0)

	If the overloading is done inside a Queue subclass, the decorator must be called without any arguments. :class:`Queue` instance is automatically provided through ``self``.

	.. code-block:: python
	   
		class MyQueue(simpype.Queue):
			def __init__(self, sim, pipe, id):
				super().__init__(sim, pipe, id)

			@simpype.queue.pop
			def pop(self):
				return self.buffer.pop(0)
	
	"""
	if isinstance(arg, simpype.Queue):
		queue = arg
		def decorator(func):
			def wrapper(queue):
				return __pop(func, queue)
			queue.pop = types.MethodType(wrapper, queue)
			return wrapper
		return decorator
	else:
		func = arg
		def wrapper(queue):
			return __pop(func, queue)
		return wrapper


class Queue:
	""" Queue is used by :class:`~simpype.pipe.Pipe` to store :class:`~simpype.message.Message` objects.

	Args:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		pipe (:class:`~simpype.pipe.Pipe`):
			The pipe this queue is associated to.
		id (str):
			The queue id.

	Attributes:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		env (simpy.Environment):
			The SimPy environment object.
		id (str):
			The queue id.
		pipe (:class:`~simpype.pipe.Pipe`):
			The pipe this queue is associated to.
		buffer (list):
			The data structure physically storing the :class:`~simpype.message.Message` objects.
		capacity (int):
			The capacity of the buffer. ``Infinite`` by default.
		active (simpy.events.Event):
			Event signaling when the queue is active.

	"""
	def __init__(self, sim, pipe, id):
		assert isinstance(sim, simpype.Simulation)
		assert isinstance(pipe, simpype.Pipe)
		self.sim = sim
		self.env = sim.env
		self.id = id
		self.pipe = pipe
		self.buffer = []
		self.capacity = float('inf')
		self.active = self.env.event().succeed()

	def _message_dropped(self, message, cause):
		assert isinstance(message, simpype.Message)
		assert message.location == self
		if message in self.buffer:
			self.buffer.remove(message)
		message.timestamp('pipe.'+str(self.id)+'.'+str(cause))

	def enable(self):
		""" Enable this queue by triggering the ``active`` attribute. """
		if not self.active.triggered:
			self.active.succeed()
		self.pipe.full()

	def disable(self):
		""" Disable this queue by resetting the ``active`` attribute. """
		self.active = self.env.event()

	def __len__(self):
		return len(self.buffer)



import simpype


class Resource(simpype.Resource):
	def __init__(self, sim, id, capacity = 1, pipe = None):
		super().__init__(sim, id, capacity, pipe)
		self.random['service'] = {0: lambda: 0}
	
	@simpype.resource.service
	def service(self, message):
		yield self.env.timeout(self.random['service'].value)


# Do NOT remove
resource = lambda *args: Resource(*args)



"""
SimPype's pipe.

"""
import inspect
import simpy
import types

import simpype
import simpype.build


def __enqueue(func, pipe, message):
	assert isinstance(pipe, Pipe)
	assert isinstance(message, simpype.Message)
	message.location = pipe
	message.resource = pipe.resource
	if inspect.isgeneratorfunction(func):
		result = yield pipe.env.process(func(pipe, message))
	else:
		result = func(pipe, message)
	pipe.full()
	return result

def __dequeue(func, pipe):
	assert isinstance(pipe, Pipe)
	if inspect.isgeneratorfunction(func):
		result = yield pipe.env.process(func(pipe))
	else:
		result = func(pipe)
	return result

def enqueue(arg):
	""" Decorator for overloading the default :class:`Pipe` enqueue behavior.

	Args:
		arg (:class:`Pipe`)(``self``):
			The :class:`Pipe` instance.
			
	If the overloading is done in scripts, the :class:`Pipe` instance must be provided as decorator argument.
	
	.. code-block:: python

		myresource = sim.add_resource(id = 'myresource')
		myresource.add_queue(id = 'myqueue')

		@simpype.pipe.enqueue(myresource.pipe)
		def enqueue(self, message):
			return self.queue['myqueue'].push(message)

	If the overloading is done inside a Pipe subclass, the decorator must be called without any arguments. :class:`Pipe` instance is automatically provided through ``self``.

	.. code-block:: python
	   
		class MyPipe(simpype.Pipe):
			def __init__(self, sim, resource, id):
				super().__init__(sim, resource, id)
				self.add_queue(id = 'myqueue')

			@simpype.pipe.enqueue
			def enqueue(self, message):
				return self.queue['myqueue'].push(message)
	
	"""
	if isinstance(arg, Pipe):
		pipe = arg
		def decorator(func):
			def wrapper(pipe, message):
				return __enqueue(func, pipe, message)
			pipe.enqueue = types.MethodType(wrapper, pipe)
			return wrapper
		return decorator
	else:
		func = arg
		def wrapper(pipe, message):
			return __enqueue(func, pipe, message)
		return wrapper

def dequeue(arg):
	""" Decorator for overloading the default :class:`Pipe` dequeue behavior.

	Args:
		arg (:class:`Pipe`)(``self``):
			The :class:`Pipe` instance.
	
			
	If the overloading is done in scripts, the :class:`Pipe` instance must be provided as decorator argument.
	
	.. code-block:: python

		myresource = sim.add_resource(id = 'myresource')
		myresource.add_queue(id = 'myqueue')

		@simpype.pipe.dequeue(myresource.pipe)
		def dequeue(self):
			return self.queue['myqueue'].pop()

	If the overloading is done inside a Pipe subclass, the decorator must be called without any arguments. :class:`Pipe` instance is automatically provided through ``self``.

	.. code-block:: python
	   
		class MyPipe(simpype.Pipe):
			def __init__(self, sim, resource, id):
				super().__init__(sim, resource, id)
				self.add_queue(id = 'myqueue')

			@simpype.pipe.dequeue
			def dequeue(self):
				return self.queue['myqueue'].pop()

	"""	
	if isinstance(arg, Pipe):
		pipe = arg
		def decorator(func):
			def wrapper(pipe):
				return __dequeue(func, pipe)
			pipe.dequeue = types.MethodType(wrapper, pipe)
			return wrapper
		return decorator
	else:
		func = arg
		def wrapper(pipe):
			return __dequeue(func, pipe)
		return wrapper


class Pipe:
	""" The pipe implements the queueing disciplines.

	Args:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		resource (:class:`~simpype.resource.Resource`):
			The resource the pipe is associated to.
		id (str):
			The pipe id.

	Attributes:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		env (simpy.Environment):
			The SimPy environment object.
		id (str):
			The pipe id.
		resource (:class:`~simpype.resource.Resource`):
			The resource the pipe is associated to.
		available (simpy.events.Event):
			Event signaling the presence of a :class:`~simpype.message.Message` in the pipe.
		queue (dict):
			The dictionary storing the :class:`~simpype.queue.Queue` instances associated to this pipe.

	"""
	def __init__(self, sim, resource, id):
		assert isinstance(sim, simpype.Simulation)
		assert isinstance(resource, simpype.Resource)
		self.sim = sim
		self.env = sim.env
		self.id = id
		self.resource = resource
		self.available = self.env.event()
		self.queue = {}
		# Init
		self.a_wait_loop = self.env.process(self._wait_loop())

	def _message_dropped(self, message, cause):
		assert isinstance(message, simpype.Message)
		assert message.location == self

	def _wait_loop(self):
		while True:
			with self.resource.use.request() as request:
				yield request & self.available
				self.available = self.env.event()
				message = yield self.env.process(self.dequeue())
				if isinstance(message, simpype.Message):
					yield self.env.process(self.resource.service(message))
				self.full()

	def add_queue(self, id, model = None):
		""" Add a new queue to the pipe.

		Args:
			id (str):
				The id of the new queue.
			model (str):
				The model of the new queue. If model is ``None``, the default queue model is created.

		Returns:
			:class:`~simpype.queue.Queue`

		"""
		queue = simpype.build.queue(self.sim, self, id, model)
		self.queue[queue.id] = queue
		return self.queue[queue.id]

	def full(self):
		""" Check if there is at least a :class:`~simpype.message.Message` in the pipe.

		Returns:
			bool

		"""
		tot = sum([len(q) for q in self.queue.values() if q.active.triggered])
		if not self.available.triggered and tot > 0:
			self.available.succeed()
			return True
		return False



"""
SimPype's message.

"""

import copy
import inspect
import simpy

import simpype
import simpype.build


class PropertyDict(dict):
	""" A custom dictionary storing simpype.Property objects.
	
	Args:
		sim (:class:`~simpype.simulation.Simulation`):
			The SimPype simulation object

	Attributes:
		sim (:class:`~simpype.simulation.Simulation`):
			The SimPype simulation object
		env (simpy.Environment): 
			The SimPy environment object

	"""
	def __init__(self, sim):
		assert isinstance(sim, simpype.Simulation)
		super().__init__()
		self.sim = sim
		self.env = sim.env
	
	def __setitem__(self, key, value):
		""" Automatically creates a simpype.Property object when assigning a value

		Args:
			key (any):
				The dictionary key
			value (any):
				The dictionaty value

		"""
		# Don't create a simpype.Property object if value is already a simpype.Property object
		if isinstance(value, Property):
			super().__setitem__(key, value)
		else:
			super().__setitem__(key, Property(self.sim, key, value))


class Property:
	""" This class implements the properties used by simpype.Message objects.
	
	A property value can be either static or dynamic. In the latter case
	the ``value`` must follow the simpype.Random value dictionary format.

	Args:
		sim (:class:`~simpype.simulation.Simulation`):
			The SimPype simulation object
		name (str):
			The property name
		value (any):
			The property value

	Attributes:
		sim (:class:`~simpype.simulation.Simulation`):
			The SimPype simulation object
		env (simpy.Environment):
			The SimPy environment object
		name (str):
			The property name
		value (any):
			The property value

	"""
	def __init__(self, sim, name, value):
		assert isinstance(sim, simpype.Simulation)
		self.sim = sim
		self.env = sim.env
		self.name = name
		# If ``value`` is a dictionary and contains lambda functions, create a simpype.Random object
		if isinstance(value, dict) and [f for f in value.values() if inspect.isfunction(f)]:
			self._random = simpype.Random(self.sim, value)
			self._value = self._random.value
		else:
			self._random = None
			self._value = value

	@property
	def value(self):
		""" The value of the simpype.Property. """
		return self._value

	def copy(self):
		""" Create a copy of the object.

		Returns:
			:class:`Property`
		"""
		property = Property(self.sim, self.name, self._value)
		property._random = self._random
		return property

	def refresh(self):
		""" Generate a new random value for the simpype.Property object. 
		
		This function does not produce any effects if the simpype.Property value is static.

		"""
		if isinstance(self._random, simpype.Random):
			self._value = self._random.value


class Timestamp:
	""" This class implements the timestamps used by simpype.Message objects.

	Timestamp objects are used for logging purposes.

	Args:
		message (:class:`Message`):
			The SimPype message object this timestamp is associated to
		timestamp (int,float):
			The simulation time of the timestmap
		resource (:class:`~simpype.resource.Resource`):
			The SimPype resource object creating the timestamp
		description (str):
			The description of the event generating the timestamp

	Attributes:
		message (:class:`Message`):
			The SimPype message object this timestamp is associated to
		timestamp (int,float):
			The simulation time of the timestmap
		resource (:class:`~simpype.resource.Resource`):
			The SimPype resource object creating the timestamp
		description (str):
			The description of the event generating the timestamp

	"""
	def __init__(self, message, timestamp, resource, description):
		assert isinstance(message, Message)
		assert isinstance(resource, simpype.Resource)
		self.message = message
		self.timestamp = float(timestamp)
		self.resource = resource
		self.description = str(description)


class Subscription:
	""" This class implements the subscriptions used by simpype.Message objects.

	Subscription used to execute a given function upong some events triggering.

	Args:
		sim (:class:`~simpype.simulation.Simulation`):
			The SimPype simulation object
		message (:class:`Message`):
			The SimPype message object this timestamp is associated to
		event (simpy.Event):
			The Simpy event the simpype.Message object is subscribed to
		callback (user-defined python function):
			The python function to call upon event triggering
		id (str):
			The simpype.Subscription id

	Attributes:
		sim (:class:`~simpype.simulation.Simulation`):
			The SimPype simulation object
		env (simpy.Environment):
			The SimPy environment object
		message (:class:`~simpype.message.Message`):
			The SimPype message object this timestamp is associated to
		event (simpy.Event):
			The Simpy event the simpype.Message object is subscribed to
		callback (user-defined python function):
			The python function to call upon event triggering
		id (str):
			The simpype.Subscription id
		disable (simpy.Event):
			The Simpy event used to remove the subscription

	"""
	def __init__(self, sim, message, event, callback, id):
		assert isinstance(sim, simpype.Simulation)
		assert isinstance(message, Message)
		assert callable(callback)
		self.sim = sim
		self.env = sim.env
		self.message = message
		self.event = event
		self.callback = callback
		self.id = id
		self.disable = self.env.event()


class Message:
	""" This class implements the simpype.Message object.

	simpype.Message object is the atomic unit processed by simpype.Resource objects.

	Args:
		sim (:class:`~simpype.simulation.Simulation`):
			The SimPype simulation object
		resource (:class:`~simpype.resource.Resource`):
			The simpype.Resource object generating this message
		id (str):
			The SimPype message id

	Attributes:
		sim (:class:`~simpype.simulation.Simulation`):
			The SimPype simulation object.
		env (simpy.Environment):
			The SimPy environment object.
		id (str):
			The simpype.Message id
		generated (float):
			The simulation time when this object was generated.
		generator (:class:`~simpype.resource.Resource`):
			The simpype.Resource that created the message.
		is_alive (bool):
			The boolean value marking if the message is alive or not. 
			A message is not alive when no further steps are available or if it is used as a template by a generator.
		location (:class:`~simpype.resource.Resource`:class:`~simpype.pipe.Pipe`:class:`~simpype.queue.Queue`):
			The location of the simpype.Message inside the simulation pipeline.
		property (:class:`PropertyDict`): 
			The PropertyDict dictionary storing the :class:`Property` objects
		seq_num (int):
			The sequence number of the message
		subscription (dict):
			The dictionary storing the Subscription objects
		visited (list):
			The list of the visited simpype.Resource objects

	"""
	def __init__(self, sim, resource, id):
		assert isinstance(sim, simpype.Simulation)
		assert isinstance(resource, simpype.Resource)
		self.sim = sim					
		self.env = sim.env
		self.id = id
		self.generated = self.env.now
		self.generator = resource
		self.is_alive = True
		self.location = resource
		self.property = PropertyDict(self.sim)
		self.seq_num = 0
		self.subscription = {}
		self.visited = []
		self.next = {}
		self.resource = resource
		self.pipeline = simpype.Pipeline(self.sim, self.id)

	@property
	def next(self):
		""" The next resources available to the message in the form of adjency list 

		Next property admits only simpype.Resource, simpype.Pipeline, and next-compatible values

		"""
		return self._next
	
	@next.setter
	def next(self, value):
		assert isinstance(value, (simpype.Resource, simpype.Pipeline, dict))
		if isinstance(value, simpype.Resource):
			self._next = {value.id: value}
		elif isinstance(value, simpype.Pipeline):
			self._next = {value.first.id: value.first}
		elif isinstance(value, dict):
			self._next = value

	@property
	def pipeline(self):
		""" The pipeline associated to the message in the form of adjency.

		A pipeline represents all the possible paths to the message.
		Pipeline property admits only simpype.Pipeline objects as a value

		"""
		return self._pipeline
	
	@pipeline.setter
	def pipeline(self, value):
		assert isinstance(value, simpype.Pipeline)
		self._pipeline = value
		if hasattr(self, 'resource'):
			self._update_next()

	@property
	def resource(self):
		""" The resource currently managing the message.

		Resource property admits only simpype.Resource objects as a value

		"""
		return self._resource
	
	@resource.setter
	def resource(self, value):
		assert isinstance(value, simpype.Resource)
		self._resource = value
		self.visited.append(value)
		if hasattr(self, 'pipeline'):
			self._update_next()

	def _drop(self, message, cause):
		""" The callback function fro dropping a message """
		self.done()
		self.location._message_dropped(self, cause)

	def _update_next(self):
		""" Update the next adjency list based on the current resource managing the message """
		if self.resource.id in self.pipeline.resource:
			self.next = {p.id: p for p in self.pipeline.resource[self.resource.id]}
		else:
			self.next = {}

	def _wait_event(self, subscription):
		""" Wait the triggering of an event and execute the associated callbak """
		value = yield subscription.event | subscription.disable
		if subscription.event in value and self.is_alive:
			subscription.callback(self, value[subscription.event])
		if subscription.id in self.subscription:
			del self.subscription[subscription.id]
	
	def copy(self):
		""" Create a dopy of this simpype.Message object. 
		
		Returns:
			:class:`Message`

		"""
		message = Message(self.sim, self.generator, self.id)
		message.generated = copy.deepcopy(self.generated)
		message.resource = self.resource
		message.location = self.location
		message.seq_num = copy.deepcopy(self.seq_num)
		message.visited = copy.copy(self.visited)
		message.is_alive = copy.deepcopy(self.is_alive)
		message.next = copy.copy(self.next)
		message.pipeline = copy.copy(self.pipeline)
		for p in self.property.values():
			message.property[p.name] = p.copy()
		for id,s in self.subscription.items():
			c = getattr(message, s.callback.__name__) if inspect.ismethod(s.callback) else s.callback
			s = message.subscribe(event = s.event, callback = c, id = id)
		return message

	def done(self):
		""" Deactivate the message, empty the next adjency list, and defuse all the active subscriptions """
		self.is_alive = False
		self.next = {}
		e_list = list(self.subscription.keys())
		for id in e_list:
			self.unsubscribe(id)

	def drop(self, id = 'dropped', event = None):
		""" Drop the simpype.Message object from the simulation.

		If event is not None, subscribe the message to the event and the default drop callback

		Args:
			id (str, optional):
				The id identifying in the log this drop action
			event (simpy.Event, optional):
				The event that will trigger the message dropping

		"""
		if event is None:
			self._drop(self, id)
		else:
			assert isinstance(event, simpy.Event)
			e = self.subscribe(event = event, callback = self._drop, id = id)

	def subscribe(self, event, callback, id):
		""" Subscribe the message to a given event which will execute a callback function.

		Args:
			event (simpy.Event):
				The simpy.Event to subscribe to
			callback (user-defined python function):
				The function to call upon event triggering
			id (str):
				The id identifying this subscription

		Returns:
			:class:`Subscription`

		"""
		assert isinstance(event, simpy.Event)
		assert callable(callback)
		if id in self.subscription:
			self.unsubscribe(id)
		s = Subscription(self.sim, self, event, callback, id)
		self.subscription[id] = s
		self.env.process(self._wait_event(s))
		return s

	def timestamp(self, description):
		""" Create and write a timestamp to the log file 
		
		Args:
			description (str):
				The timestamp description
		
		"""
		ts = Timestamp(self, self.env.now, self.resource, description)
		self.sim.log.write(ts)
		return ts

	def unsubscribe(self, id):
		""" Unsubscribe the message from the subscription ``id``

		Args:
			id (any):
				The subscription id to unsubscribe

		"""
		assert id in self.subscription
		self.subscription[id].disable.succeed()



"""
The ``simpype`` module aggregates SimPype's most used components into a single
namespace. This is purely for convenience. You can of course also access
everything (and more!) via their actual submodules.

The following tables list all of the available components in this module.

{toc}

"""
from pkgutil import extend_path

from simpype.simulation import Simulation
from simpype.message import Message
from simpype.pipe import Pipe
from simpype.pipeline import Pipeline
from simpype.queue import Queue
from simpype.random import Random
from simpype.resource import Resource


def compile_toc(entries, section_marker='='):
	"""Compiles a list of sections with objects into sphinx formatted
	autosummary directives."""
	toc = ''
	for section, objs in entries:
		toc += '\n\n%s\n%s\n\n' % (section, section_marker * len(section))
		toc += '.. autosummary::\n\n'
		for obj in objs:
			toc += '	~%s.%s\n' % (obj.__module__, obj.__name__)
	return toc


toc = (
	('Simulation', (
		Simulation,
	)),
	('Resource', (
		Resource,
	)),
	('Pipe', (
		Pipe,
	)),
	('Pipeline', (
		Pipeline,
	)),
	('Queue', (
		Queue,
	)),
	('Message', (
		Message,
	)),
)

# Use the toc to keep the documentation and the implementation in sync.
if __doc__:
	__doc__ = __doc__.format(toc=compile_toc(toc))
__all__ = [obj.__name__ for section, objs in toc for obj in objs]

__path__ = extend_path(__path__, __name__)
__version__ = '1.0.0.rc6'



"""
SimPype's build library.

"""

import importlib
import logging
import os


def _import(model, prefix):
	""" Dynamically import files given a path and a model.

	This functions imports a python source file as python model.

	Args:
		model (str):
			The name of the file to import withouth .py extension
		prefix (str):
			The path where the model is stored

	Returns:
		user-defined python module
	"""
	module = importlib.machinery.SourceFileLoader(model, os.path.join(prefix, model+'.py')).load_module()
	return module

def logger(name, path):
	""" Create a logger object.

	This function creates a logger object for logging the simulation results.
	A logger accepts a (str) as input and writes it to the log file.

	Args:
		name (str):
			The name of the logger
		path (str):
			The path of the log file managed by the logger

	Returns:
		Python Logger from logging module :class:`logging.Logger`
	"""
	# Create a logger
	logger = logging.getLogger(path)
	logger.setLevel(logging.INFO)
	# Create a file handler
	# delay = True does not create the logging file if nothing is written
	fh = logging.FileHandler(path, mode = 'w', delay = True)
	# Define the format of the messages to be logged
	formatter = logging.Formatter('%(message)s')
	fh.setFormatter(formatter)
	# Add the file handler to the logger
	logger.addHandler(fh)
	return logger

def resource(sim, id, model = None, capacity = 1, pipe = None):
	""" Create a :class:`~simpype.resource.Resource` object.

	Dynamically build a resource object based on the model.
	If model is None the default resource model is built.

	Args:
		sim (:class:`~simpype.simulation.Simulation`): 
			The SimPype simulation object
		id (str):
			The resource id to build
		model (str, optional):
			The model of the resource
		capacity (int, optional):
			The capacity of the resource
		pipe (str, optional):
			The model of the pipe associated to the resource

	Returns:
		:class:`~simpype.resource.Resource`
	"""
	if model is None:
		module = importlib.import_module('simpype.model.resource')
	else:
		module = _import(model, sim.model.dir)
	return module.resource(sim, id, capacity, pipe)

def generator(sim, id, model = None):
	""" Create a :class:`~simpype.resource.Resource` object implementing a generator.

	Dynamically build a generator object based on the model.
	If model is None the default generator model is built.

	Args:
		sim (:class:`~simpype.simulation.Simulation`): 
			The SimPype simulation object
		id (str):
			The resource id to build
		model (str, optional):
			The model of the resource

	Returns:
		:class:`~simpype.resource.Resource`
	"""
	if model is None:
		module = importlib.import_module('simpype.model.generator')
	else:
		module = _import(model, sim.model.dir)
	return module.resource(sim, id)

def pipe(sim, resource, id, model = None):
	""" Create a :class:`~simpype.pipe.Pipe` object.

	Dynamically build a pipe object based on the model.
	If model is None the default pipe model is built.

	Args:
		sim (:class:`~simpype.simulation.Simulation`): 
			The SimPype simulation object
		resource (:class:`~simpype.resource.Resource`):
			The resource object this pipe is associated to
		id (str):
			The pipe id to build
		model (str, optional):
			The model of the pipe

	Returns:
		:class:`~simpype.pipe.Pipe`
	"""
	if model is None:
		module = importlib.import_module('simpype.model.pipe')
	else:
		module = _import(model, sim.model.dir)
	return module.pipe(sim, resource, id)

def queue(sim, pipe, id, model = None):
	""" Create a :class:`~simpype.queue.Queue` object.

	Dynamically build a queue object based on the model.
	If model is None the default queue model is built.

	Args:
		sim (:class:`~simpype.simulation.Simulation`): 
			The SimPype simulation object
		pipe (:class:`~simpype.pipe.Pipe`):
			The pipe object this queue is associated to
		id (str):
			The queue id to build
		model (str, optional):
			The model of the queue

	Returns:
		:class:`~simpype.queue.Queue`
	"""
	if model is None:
		module = importlib.import_module('simpype.model.queue')
	else:
		module = _import(model, sim.model.dir)
	return module.queue(sim, pipe, id)



# Always prefer setuptools over distutils
from setuptools import setup
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='simpype',
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.0.0.rc6',
    description='SimPype is a simulation framework based on Simpy that relies on the concepts of resource and pipe.',
    long_description=long_description,
    # The project's main homepage.
    url='https://github.com/Mallets/SimPype/',
    # Author details
    author='Luca Cominardi',
    author_email='luca.cominardi@gmail.com',
    # Choose your license
    license='GPLv3+ License',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
		'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
	python_requires='>=3.3',
    keywords='simulation queue pipe simpy',
	packages=['simpype', 'simpype.model'],
    install_requires=['simpy>=3.0.10'],
)



"""
SimPype's resources process :class:`~simpype.message.Message` objects.
The behavior of each resource can be customized by overloading the custom
``service`` function through the decorator ``@simpype.resource.service``.
See :ref:`resource` :func:`service` for more details on how to customize a :class:`Resource` behavior.

.. code-block :: python

   import simpype
   import random

   sim = simpype.Simulation(id = 'simple')
   gen0 = sim.add_generator(id = 'gen0')
   gen0.message.property['wait'] = {
       0: lambda: random.uniform(0,1)
   }
   res0 = sim.add_resource(id = 'res0')
   res0.random['service'] = {
       0: lambda: 2.0
   }

   @simpype.resource.service(res0)
   def service(self, message):
       # Wait for a random time
       yield self.env.timeout(self.random['service'])
       # Wait for a time as reported in the message property
       yield self.env.timeout(message.property['wait'].value)

   sim.run(until = 10)

"""
import inspect
import simpy
import types

import simpype
import simpype.build


def __service(func, resource, message):
	assert isinstance(resource, Resource)
	assert isinstance(message, simpype.Message)
	message.location = resource
	if inspect.isgeneratorfunction(func):
		mid = str(message.id)+str(message.seq_num)
		a_serve = resource.env.process(func(resource, message))
		resource.task[mid] = Task(message.sim, message, a_serve)
		try:
			yield a_serve
			message.timestamp('resource.serve')
		except simpy.Interrupt as interrupt:
			message.timestamp('resource.'+str(interrupt.cause))
		del resource.task[mid]
	else:
		func(resource, message)
		message.timestamp('resource.serve')
	if message.next:
		resource.send(message)
	else:
		message.done()


def service(arg):
	""" Decorator for overloading the default :class:`Resource` service behavior.


	Args:
		arg (:class:`Resource`)(``self``):
			The :class:`Resource` instance.
	
			
	If the overloading is done in scripts, the :class:`Resource` instance must be provided as decorator argument.
	
	.. code-block:: python

		myresource = sim.add_resource(id = 'myresource1')

		@simpype.resource.service(myresource)
		def service(self, message):
			yield self.env.timeout(1.0)

	If the overloading is done inside a Resource subclass, the decorator must be called without any arguments. :class:`Resource` instance is automatically provided through ``self``.

	.. code-block:: python
	   
		class MyResource(simpype.Resource):
			def __init__(self, sim, id, capacity = 1, pipe = None):
				super().__init__(sim, id, capacity, pipe)

			@simpype.resource.service
			def service(self, message):
				yield self.env.timeout(1.0)
	
	"""
	if isinstance(arg, simpype.Resource):
		resource = arg
		def decorator(func):
			def wrapper(resource, message):
				return __service(func, resource, message)
			resource.service = types.MethodType(wrapper, resource)
			return wrapper
		return decorator
	else:
		func = arg
		def wrapper(resource, message):
			return __service(func, resource, message)
		return wrapper


class Task:
	""" This class implements the :class:`Task` managed by the :class:`Resource`.

	Args:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		message (:class:`~simpype.message.Message`):
			The message being processed in this class
		process (simpy.events.Process):
			The SimPy process to execute

	Attributes:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		env (simpy.Environment):
			The SimPy environment object.
		message (:class:`~simpype.message.Message`):
			The message being processed in this class
		started (float):
			The simulation time this task was started
		interrupted (float):
			The simulation time this task was interrupted. ``None`` if active.
		process (simpy.events.Process):
			The SimPy process being executed

	"""
	def __init__(self, sim, message, process):
		assert isinstance(sim, simpype.Simulation)
		self.sim = sim
		self.env = sim.env
		self.message = message
		self.started = self.env.now
		self.interrupted = None
		self.process = process

	def interrupt(self, cause = None):
		self.process.interrupt(cause = cause)
		self.interrupted = self.env.now


class Resource:
	""" This class implements the :class:`Resource` object.

	Args:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		id (str):
			The resource id.
		capacity (int):
			The simpy.Resource capacity.
		pipe (:class:`Pipe`):
			The SimPype pipe model associated to this resource.

	Attributes:
		sim (:class:`Simulation`):
			The SimPype simulation object.
		env (simpy.Environment):
			The SimPy environment object.
		id (str):
			The simpype.Resource id.
		use (simpy.Resource):
			The SimPy resource object.
		pipe (:class:`Pipe`):
			The SimPype pipe object.
		random (:class:`RandomDict`):
			The SimPype RandomDict object.
		task (dict):
			The dictionary storing the task currenty being executed by the resource.

	"""
	def __init__(self, sim, id, capacity = 1, pipe = None):
		assert isinstance(sim, simpype.Simulation)
		self.sim = sim
		self.env = sim.env
		self.id = id
		self.use = simpy.Resource(self.env, capacity = capacity)
		self.pipe = simpype.build.pipe(self.sim, self, self.id, pipe)
		self.random = simpype.random.RandomDict(self.sim)
		self.task = {}

	def _message_dropped(self, message, cause):
		assert isinstance(message, simpype.Message)
		assert message.location == self
		mid = message.id+str(message.seq_num)
		assert mid in self.task
		self.task[mid].interrupt(cause = cause)

	def send(self, message):
		""" Send a :class:`~simpype.message.Message`.

		Args:
			message (:class:`~simpype.message.Message`):
				The message to send.

		"""
		assert isinstance(message, simpype.Message)
		resource = list(message.next.values())
		for i in range(0, len(resource)):
			tmsg = message if i == (len(resource)-1) else message.copy()
			self.env.process(resource[i].pipe.enqueue(tmsg))



import simpype


class Generator(simpype.Resource):
	def __init__(self, sim, id):
		super().__init__(sim, id)
		self.message = simpype.Message(self.sim, self, self.id)
		self.counter = 0
		self.to_send = float("inf")
		# Init
		self.message.is_alive = False
		self.random['arrival'] = {0: lambda: float("inf")}
		self.a_gen = self.env.process(self.h_gen())

	def gen_message(self):
		message = self.message.copy()
		message.seq_num = self.counter
		message.generated = self.env.now
		for p in message.property.values():
			p.refresh()
		if 'lifetime' in message.property:
			message.drop('lifetime', self.env.timeout(message.property['lifetime'].value, 'expired'))
		message.is_alive = True
		return message

	def h_gen(self):
		more = True
		while self.counter < self.to_send and more:
			val = self.random['arrival'].value
			if val is None:
				more = False
			else:
				yield self.env.timeout(val)
				message = self.gen_message()
				self.send(message)
				self.counter = self.counter + 1

# Do NOT remove
resource = lambda *args: Generator(*args)
